from typing import Tuple

import numpy as np
from scipy.optimize import minimize
from scipy.spatial.distance import pdist, squareform

from .base import NCAFS


def exponential_kernel(z: np.ndarray, sigma: float) -> np.ndarray:
    """
    Exponential kernel.

    Parameters
    ----------
    z : array
        Input array
    sigma : float
        The length scale; must be a positive float.

    Returns
    -------
    array
        Transformed values.
    """
    return np.exp(-z / sigma)


def pairwise_feature_distance(X: np.ndarray, metric: str = 'cityblock') -> np.ndarray:
    """
    Calculates pairwise distances between each sample in each feature.

    Parameters
    ----------
    X : numpy.ndarray
        An (N x M) data matrix where N is the number of samples and M is the number of features.
    metric : str (optional, default='cityblock')
        Distance metric to use.

    Returns
    -------
    array
        An (M x N X N) numpy array where array[0] is the distance matrix
        representing pairwise distances between samples in feature 0.
    """
    # matrix to hold pairwise distances between samples in each feature
    N, M = X.shape
    dists = np.zeros((M, N, N))
    for j in range(M):
        dists[j] = squareform(
            pdist(
                X[:, j].reshape(-1, 1),
                metric=metric
            )
        )
    return dists


def add_pseudocount(v: np.ndarray) -> np.ndarray:
    """
    Adds pseudocount to avoid zero-value elements in a vector.

    Parameters
    ----------
    v : array of shape (n, 1)
        Vector that may contain undesired zeros.

    Returns
    -------
    array of shape (n, 1)
        The original vector with additive smoothing such that for every element `v_i > 0`
    """
    v = v.copy()
    mask = v == 0
    n_zeros = np.sum(mask)
    if n_zeros > 0:
        if n_zeros == len(v):
            pseudocount = np.exp(-20)
        else:
            pseudocount = np.min(v[~mask])
        v[mask] += pseudocount
    return v


def cost_nca(
        w: np.ndarray,
        feat_dist: np.ndarray,
        y_loss: np.ndarray,
        sigma: float = 1.0,
        alpha: float = 1.0
) -> Tuple[float, np.ndarray]:
    """
    Cost function for NCA model.

    Parameters
    ----------
    w : array of shape (n_features,)
        weights array
    feat_dist : array of shape (n_features, n_samples, n_samples)
        Feature pairwise distance array
    y_loss : array of shape (n_samples, n_samples)
        Target values pairwise distance (loss)
    sigma : float (optional, default=1)
        The length scale of the exponential kernel; must be a positive float.
    alpha : float (optional, default=0.01)
        Regularization strength; must be a positive float.

    Returns
    -------
    cost : float
        Value of the cost function at `w`.
    gradient : array of shape (n_features,)
        Gradient vector at `w`.
    """
    M, N, _ = feat_dist.shape
    w = w.reshape((M, 1, 1))

    Dw = np.sum(feat_dist * w ** 2, axis=0)
    kDw = exponential_kernel(Dw, sigma)

    np.fill_diagonal(kDw, 0)
    row_sum = np.sum(kDw, axis=1, keepdims=True)
    row_sum = add_pseudocount(row_sum)
    p = kDw / row_sum
    # assert np.allclose(p.sum(axis=1), 1)

    E_loss = np.sum(p * y_loss, axis=1, keepdims=True)

    w = w.reshape(-1, 1)
    cost = np.mean(E_loss) + alpha * np.dot(w.T, w).item()

    s1 = np.sum(E_loss * p * feat_dist, axis=(1, 2)).reshape(-1, 1)
    s2 = np.sum(y_loss * p * feat_dist, axis=(1, 2)).reshape(-1, 1)
    grad = 2 * ((1 / sigma) * (s1 - s2) * (1 / N) + alpha) * w
    grad = grad.reshape(-1)

    return cost, grad


class NCAFSR(NCAFS):
    """
    Neighborhood Component Analysis Feature Selection for regression targets.
    """

    def _fit(self, X: np.ndarray, y: np.ndarray, w0: np.ndarray) -> np.ndarray:
        n_feat = X.shape[1]
        X_dist = pairwise_feature_distance(X, metric=self.metric)
        y_dist = squareform(pdist(y.reshape(-1, 1), metric=self.metric))

        if self.alpha is None:
            alpha = self._estimate_alpha(n_feat, y)
        else:
            alpha = self.alpha
        args = (X_dist, y_dist, self.sigma, alpha)
        result = minimize(
            cost_nca, w0, args,
            method=self.solver,
            jac=True,
            options=dict(disp=False),
            bounds=[self.bounds] * n_feat
        )
        return np.abs(result['x'])

    @staticmethod
    def _estimate_alpha(n_feat: int, y: np.ndarray) -> float:
        return 0.01 * np.log10(n_feat) * y.std()


class NCAFSC(NCAFS):
    """
    Neighborhood Component Analysis Feature Selection for classification targets.
    """

    def _fit(self, X: np.ndarray, y: np.ndarray, w0: np.ndarray) -> np.ndarray:
        n_feat = X.shape[1]
        X_dist = pairwise_feature_distance(X, metric=self.metric)
        y_dist = squareform(pdist(y.reshape(-1, 1), metric=lambda u, v: 1 - (u == v)))

        if self.alpha is None:
            alpha = self._estimate_alpha(n_feat, y)
        else:
            alpha = self.alpha
        args = (X_dist, y_dist, self.sigma, alpha)
        result = minimize(
            cost_nca, w0, args,
            method=self.solver,
            jac=True,
            options=dict(disp=False),
            bounds=[self.bounds] * n_feat
        )
        return np.abs(result['x'])

    @staticmethod
    def _estimate_alpha(n_feat: int, y: np.ndarray) -> float:
        return 0.01 / max(1, np.log10(n_feat))



import pytest
from sklearn import datasets

from ncafs import NCAFSR, NCAFSC


eps = 1e-3


class TestRegression:
    @pytest.mark.parametrize("n_features, n_informative", [(20, 5), (20, 10), (20, 15), (40, 10), (40, 20)])
    def test_toy_dataset(self, n_features, n_informative):
        X, y, coef = datasets.make_regression(
            n_samples=1000,
            n_features=n_features,
            n_informative=n_informative,
            bias=0,
            noise=1e-3,
            coef=True,
            shuffle=False,
            random_state=0
        )

        fs_reg = NCAFSR()
        fs_reg.fit(X, y)
        w = fs_reg.weights_
        n_select = len(w[w > eps])
        assert 0.8 * n_informative - 0.1 * n_features <= n_select <= 1.2 * n_informative + 0.1 * n_features

    @pytest.mark.parametrize("n_features", [10, 20, 40, 60, 100])
    def test_friedman1(self, n_features):
        X, y = datasets.make_friedman1(n_samples=1000, n_features=n_features, noise=1e-1)
        fs_reg = NCAFSR()
        fs_reg.fit(X, y)
        w = fs_reg.weights_
        assert len(w[w > 1e-3]) == 5

    @pytest.mark.parametrize("n_features", [10, 20, 40, 60, 100])
    def test_sparse_uncorrelated(self, n_features):
        X, y = datasets.make_sparse_uncorrelated(n_samples=1000, n_features=n_features)
        fs_reg = NCAFSR()
        fs_reg.fit(X, y)
        w = fs_reg.weights_
        assert len(w[w > eps]) == 4


class TestClassification:
    @pytest.mark.parametrize("n_features, n_informative", [(20, 5), (20, 10), (20, 15), (40, 10), (40, 20),
                                                           (100, 10), (100, 40)])
    def test_toy_dataset(self, n_features, n_informative, n_redundant=0, n_repeated=0):
        X, y = datasets.make_classification(
            n_samples=1000,
            n_classes=5,
            n_features=n_features,
            n_informative=n_informative,
            n_redundant=n_redundant,
            n_repeated=n_repeated,
            flip_y=0.1,
            class_sep=0.5,
            shuffle=False,
            random_state=0
        )
        fs_class = NCAFSC()
        fs_class.fit(X, y)
        w = fs_class.weights_
        n_select = len(w[w > eps])
        assert 0.8*n_informative <= n_select <= 1.2*n_informative



# https://scikit-learn.org/stable/developers/develop.html
# https://github.com/scikit-learn-contrib/project-template/blob/master/skltemplate/_template.py

from abc import ABC, abstractmethod
from typing import Optional, Tuple

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import KFold
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import (
    check_X_y,
    check_array,
    check_is_fitted,
    check_scalar
)


class NCAFS(TransformerMixin, BaseEstimator, ABC):
    """

    Parameters
    ----------
    n_features_to_select : int (optional, default=None)
        Number of features which must be selected.
    threshold : float (optional, default=0.1)
        Used only when `n_features_to_select=None`. Select features whose weight is greater then `threshold`.
    alpha : float (optional, default=None)
        Regularization strength; must be a positive float. If None, then it is automatically set based on an heuristic.
    sigma : float (optional, default=1)
        The length scale of the exponential kernel; must be a positive float.
    metric : str (optional, default='cityblock')
        The metric used to calculate pairwise feature distances.
    standardize : bool (optional, default=True)
        Whether to standardize feature values.
    fit_method : str (optional, default='auto')
        If 'full', solves the optimization problem using the whole dataset at once. If 'average', splits the data into
        folds and returns the average of the feature weights by fold. If 'auto', determines the best number of folds
        to achieve a good trade-off between performance and information.
    n_splits : int (optional, default=3)
        Number of folds to split the data into when `fit_method='average'`.
    solver : str (optional, default='L-BFGS-B')
        The solver method used for minimization of the NCA cost function.
    """

    def __init__(
            self,
            n_features_to_select: Optional[int] = None,
            threshold: float = 1e-1,
            alpha: float = None,
            sigma: float = 1.0,
            metric: str = 'cityblock',
            standardize: bool = True,
            fit_method: str = 'auto',
            n_splits: int = 3,
            solver='L-BFGS-B',
            bounds: Tuple[float, float] = (0, 10),
    ):
        check_scalar(threshold, 'threshold', (float, int), min_val=0)
        check_scalar(sigma, 'sigma', (float, int), min_val=1e-7)
        check_scalar(n_splits, 'n_split', int, min_val=2)

        if alpha is not None:
            check_scalar(alpha, 'alpha', (float, int), min_val=0)

        if fit_method not in ['full', 'average', 'auto']:
            ValueError(f"Unknown fit method '{fit_method}'. It must be either 'full', 'average' or 'auto'.")

        self.n_features_to_select = n_features_to_select
        self.threshold = threshold
        self.alpha = alpha
        self.sigma = sigma
        self.metric = metric
        self.standardize = standardize
        self.fit_method = fit_method
        self.n_splits = n_splits
        self.solver = solver
        self.bounds = bounds

    @abstractmethod
    def _fit(self, X: np.ndarray, y: np.ndarray, w0: np.ndarray) -> np.ndarray:
        pass

    @staticmethod
    @abstractmethod
    def _estimate_alpha(n_feat: int, y: np.ndarray) -> float:
        pass

    def fit(self, X, y, w0=None):
        """
        Fit the NCAFS model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The training input samples.
        y : array-like of shape (n_samples,)
            The target values.
        w0 : array of shape (n_features,)
            The initial weight values.
        """
        X_, y_ = check_X_y(X, y, accept_sparse=False, y_numeric=True)
        y_ = y_.reshape(-1, 1)
        n_inst, n_feat = X_.shape

        if self.standardize:
            scaler = StandardScaler()
            X_ = scaler.fit_transform(X_)

        if w0 is None:
            w0 = np.ones(n_feat)

        if self.fit_method == 'auto':
            if n_inst > 1e3:
                n_splits = n_inst // 500
                fit_method = 'average'
            else:
                n_splits = self.n_splits
                fit_method = 'full'
        else:
            fit_method = self.fit_method
            n_splits = self.n_splits

        if fit_method == 'full':
            w = self._fit(X_, y_, w0)
        else:
            w_fold = []
            kf = KFold(n_splits=n_splits, shuffle=False)
            for _, index in kf.split(X_):
                w_fold.append(self._fit(X_[index], y_[index], w0))
            w_fold = np.array(w_fold)
            w = w_fold.mean(axis=0)

        n_feat_select = self.n_features_to_select
        if n_feat_select is None:
            self.support_ = (w > self.threshold)
            self.n_features_ = np.sum(self.support_)
        elif isinstance(n_feat_select, int):
            if 0 < n_feat_select < n_feat:
                first_n_feat_ind = np.argsort(w)[::-1][:n_feat_select]
                support = np.zeros(w.shape, dtype=bool)
                support[first_n_feat_ind] = True
                self.support_ = support
                assert np.sum(support) == n_feat_select
                self.n_features_ = n_feat_select
            else:
                raise ValueError("n_features_to_select must be greater than zero and less than the number of features.")
        else:
            raise TypeError("n_features_to_select must be either None or integer.")

        self.weights_ = w
        self.X_ = X_
        self.y_ = y_

        return self

    def transform(self, X) -> np.ndarray:
        """
        Reduce X to the selected features.

        Parameters
        ----------
        X : array of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        X_r : array of shape (n_samples, n_selected_features)
            The input samples with only the selected features.
        """
        check_is_fitted(self, ['weights_', 'support_', 'n_features_'])
        X_ = check_array(X)
        return X_[:, self.support_]



import numpy as np
import pytest

from ncafs.core import add_pseudocount


class TestFunctions:
    @pytest.mark.parametrize("length, frac_zeros", [(10, 0.1), (10, 0.5), (10, 1), (100, 0), (100, 0.1), (100, 0.8)])
    def test_pseudocount(self, length, frac_zeros):
        A = np.random.rand(length, 1) + 0.5
        idx = np.arange(length)
        to_zero = np.random.choice(idx, size=round(length * frac_zeros), replace=False)
        A[to_zero] = 0
        A_pseudo = add_pseudocount(A)

        assert np.sum(A > 0) == round(length * (1 - frac_zeros))
        assert (A_pseudo > 0).all()
        assert A.shape == A_pseudo.shape
        assert np.alltrue(A[A > 0] == A_pseudo[A > 0])
